import numpy as np

from gradman import Tensor


class Module:
    def _forward(self, i):
        return self.forward(i)

    __call__ = _forward

    def parameters(self):
        return [getattr(self, i) for i in self.__dict__]


class Linear(Module):
    def __init__(self, idim: int, odim: int):
        super(Linear, self).__init__()

        self.idim = idim
        self.odim = odim

        self.w = Tensor(np.random.randn(idim, odim), requires_grad=True)
        self.b = Tensor(np.random.randn(), requires_grad=True)

    def forward(self, i: Tensor):
        return (i @ self.w) + self.b

    __call__ = forward

    def parameters(self):
        return [self.w, self.b]

    def update(self, p_new):
        self.w, self.b = p_new



from gradman import Tensor


class Optimizer:
    def step(self, params):

        for layer in params:
            if not isinstance(layer.parameters()[-1], Tensor):
                self.step(layer.parameters())
            else:
                p_new = [self.f(i) for i in layer.parameters()]
                layer.update(p_new)


class GDE(Optimizer):
    def __init__(self, lr: float = 0.001):
        super(GDE, self).__init__()

        self.lr = Tensor(lr)

    def f(self, p):
        return p - (self.lr * p.grad)



from typing import List, Optional, Union

import numpy as np

from gradman.ctg import ContextGraph
from gradman.ops import add_, matmul_, mul_, neg_, pow_, slice_, sum_


class Tensor:

    Tensorable = Union[int, float, list, np.ndarray]

    def __init__(
        self,
        data: Tensorable,
        requires_grad: bool = False,
        _ctx: List[ContextGraph] = None,
    ) -> None:

        self._data = self.UnTensored(data)
        self.requires_grad = requires_grad
        self._ctx = _ctx or []
        self.grad: Optional["Tensor"] = None
        self.shape = self.data.shape

        if self.requires_grad:
            self.zero_grad()

    @property
    def data(self) -> np.ndarray:
        return self._data

    @data.setter
    def data(self, t: np.ndarray):
        self._data = t
        self.grad = None

    def UnTensored(self, t: Tensorable) -> np.ndarray:
        if isinstance(t, np.ndarray):
            return t
        else:
            return np.array(t)

    def __repr__(self) -> str:
        return f"<Tensor ({self.data}, requires_grad={self.requires_grad})>"

    def __getitem__(self, idx) -> "Tensor":
        o, requires_grad, _ctx = slice_(self, idx)
        return Tensor(o, requires_grad, _ctx)

    def zero_grad(self) -> None:
        self.grad = Tensor(np.zeros_like(self.data, dtype=np.float64))

    def backward(self, grad: "Tensor" = None) -> None:
        assert self.requires_grad, "Called backward() on non-requires-grad Tensor"

        if grad is None:
            if self.shape == ():
                grad = Tensor(1.0)
            else:
                raise RuntimeError("`grad` must be specified for non 0 tensor")

        self.grad.data += grad.data

        for c in self._ctx:
            c.tensor.backward(Tensor(c.grad_fn(grad.data)))

    def sum(self) -> "Tensor":
        """Sum of elements of a tensor"""
        return Tensor(*sum_(self))

    def add(t1: "Tensor", t2: "Tensor") -> "Tensor":
        """Addition of two tensors"""
        return Tensor(*add_(t1, t2))

    def mul(t1: "Tensor", t2: "Tensor") -> "Tensor":
        """Multiplication of two tensors"""
        return Tensor(*mul_(t1, t2))

    def neg(self) -> "Tensor":
        """Negative of a tensor"""
        return Tensor(*neg_(self))

    def sub(t1: "Tensor", t2: "Tensor") -> "Tensor":
        """Subtraction of two tensors"""
        return t1 + (-t2)

    def matmul(t1: "Tensor", t2: "Tensor") -> "Tensor":
        """Tensor Matrix Multiplication"""
        return Tensor(*matmul_(t1, t2))

    def pow(self, p: float) -> "Tensor":
        """Tensor power"""
        return Tensor(*pow_(self, p))

    __add__ = add
    __mul__ = mul
    __neg__ = neg
    __sub__ = sub
    __matmul__ = matmul
    __pow__ = pow



from typing import List, Tuple

import numpy as np

from gradman.ctg import ContextGraph

"""Collection of all Tensor supported operations"""


def sum_(t: "Tensor") -> Tuple[np.ndarray, bool, List[ContextGraph]]:
    """Sum of elements of a Tensor

    Parameters
    ----------
    t : 'Tensor'
        t

    Returns
    -------
    Tuple[np.ndarray, bool, List[ContextGraph]]

    """
    o = t.data.sum()
    requires_grad = t.requires_grad

    if requires_grad:

        def grad_fn(grad: np.ndarray) -> np.ndarray:
            return grad * np.ones_like(t.data)

        _ctx = [ContextGraph(t, grad_fn)]
    else:
        _ctx = []

    return o, requires_grad, _ctx


def pow_(t: "Tensor", p: float) -> Tuple[np.ndarray, bool, List[ContextGraph]]:
    """Tensor to the power `p`"""
    o = t.data ** p
    requires_grad = t.requires_grad

    if requires_grad:

        def grad_fn(grad: np.ndarray) -> np.ndarray:
            return grad * (t.data ** (p - 1)) * p

        _ctx = [ContextGraph(t, grad_fn)]
    else:
        _ctx = []

    return o, requires_grad, _ctx


def add_(t1: "Tensor", t2: "Tensor") -> Tuple[np.ndarray, bool, List[ContextGraph]]:
    """Addition of two tensors

    Parameters
    ----------
    t1 : "Tensor"
        t1
    t2 : 'Tensor'
        t2

    Returns
    -------
    Tuple[np.ndarray, bool, List[ContextGraph]]

    """

    o = t1.data + t2.data

    requires_grad = t1.requires_grad or t2.requires_grad
    _ctx: List[ContextGraph] = []

    if t1.requires_grad:

        def grad_fn1(grad: np.ndarray) -> np.ndarray:

            for _ in range(grad.ndim - t1.data.ndim):
                grad = grad.sum(axis=0)

            for i, dim in enumerate(t1.shape):
                if dim == 1:
                    grad = grad.sum(axis=i, keepdims=True)

            return grad

        _ctx.append(ContextGraph(t1, grad_fn1))

    if t2.requires_grad:

        def grad_fn2(grad: np.ndarray) -> np.ndarray:

            for _ in range(grad.ndim - t2.data.ndim):
                grad = grad.sum(axis=0)

            for i, dim in enumerate(t2.shape):
                if dim == 1:
                    grad = grad.sum(axis=i, keepdims=True)

            return grad

        _ctx.append(ContextGraph(t2, grad_fn2))

    return o, requires_grad, _ctx


def mul_(t1: "Tensor", t2: "Tensor") -> Tuple[np.ndarray, bool, List[ContextGraph]]:
    o = t1.data * t2.data

    requires_grad = t1.requires_grad or t2.requires_grad
    _ctx: List[ContextGraph] = []

    if t1.requires_grad:

        def grad_fn1(grad: np.ndarray) -> np.ndarray:
            grad = grad * t2.data

            for _ in range(grad.ndim - t1.data.ndim):
                grad = grad.sum(axis=0)

            for i, dim in enumerate(t1.shape):
                if dim == 1:
                    grad = grad.sum(axis=i, keepdims=True)

            return grad

        _ctx.append(ContextGraph(t1, grad_fn1))

    if t2.requires_grad:

        def grad_fn2(grad: np.ndarray) -> np.ndarray:
            grad = grad * t1.data

            for _ in range(grad.ndim - t2.data.ndim):
                grad = grad.sum(axis=0)

            for i, dim in enumerate(t2.shape):
                if dim == 1:
                    grad = grad.sum(axis=i, keepdims=True)

            return grad

        _ctx.append(ContextGraph(t2, grad_fn2))

    return o, requires_grad, _ctx


def neg_(t: "Tensor") -> Tuple[np.ndarray, bool, List[ContextGraph]]:

    o = -t.data
    requires_grad = t.requires_grad
    if requires_grad:
        _ctx = [ContextGraph(t, lambda x: -x)]
    else:
        _ctx = []

    return o, requires_grad, _ctx


def matmul_(t1: "Tensor", t2: "Tensor") -> Tuple[np.ndarray, bool, List[ContextGraph]]:

    o = t1.data @ t2.data

    requires_grad = t1.requires_grad or t2.requires_grad
    _ctx: List[ContextGraph] = []

    if t1.requires_grad:

        def grad_fn1(grad: np.ndarray) -> np.ndarray:
            return grad @ t2.data.T

        _ctx.append(ContextGraph(t1, grad_fn1))

    if t2.requires_grad:

        def grad_fn2(grad: np.ndarray) -> np.ndarray:
            return t1.data.T @ grad

        _ctx.append(ContextGraph(t2, grad_fn2))

    return o, requires_grad, _ctx


def slice_(t: "Tensor", idx) -> Tuple[np.ndarray, bool, List[ContextGraph]]:

    o = t.data[idx]
    requires_grad = t.requires_grad

    if requires_grad:

        def grad_fn(grad: np.ndarray) -> np.ndarray:
            partial_grad = np.zeros_like(o)
            partial_grad[idx] = grad
            return partial_grad

        _ctx.append(ContextGraph(t, grad_fn))

    else:
        _ctx = []

    return o, requires_grad, _ctx
